@@ -21,14 +21,13 @@ our @EXPORT = qw( _sort_entries
 		  _hexdump
 		  $debug
                 );
-our @EXPORT_OK = qw( _do_nothing
+our @EXPORT_OK = qw( _is_lnk
+                     _is_dir
+                     _is_reg
+                     _do_nothing
 		     _glob_to_regex
                      _tcroak );
 
-
-
-
-
 our $debug;
 
 BEGIN {
@@ -175,7 +174,6 @@ sub _glob_to_regex {
 		    else {
 			croak "invalid glob pattern";
 		    }
-				
 		}
 		else {
                     $wildcards--;
@@ -208,7 +206,7 @@ sub _catch_tainted_args {
     my $i;
     for (@_) {
         next unless $i++;
-        if (tainted $_) {
+        if (tainted($_)) {
             my (undef, undef, undef, $subn) = caller 1;
             my $msg = ( $subn =~ /::([a-z]\w*)$/
                         ? "Insecure argument '$_' on '$1' method call"
@@ -216,30 +214,29 @@ sub _catch_tainted_args {
             _tcroak($msg);
         }
         elsif (ref($_)) {
-            local $@;
-            local $SIG{__DIE__};
-            for (eval { values %$_ }) {
-                if (tainted $_) {
-                    my (undef, undef, undef, $subn) = caller 1;
-                    my $msg = ( $subn =~ /::([a-z]\w*)$/
-                                ? "Insecure argument on '$1' method call"
-                                : "Insecure argument on method call" );
-                    _tcroak($msg);
-                }
+            for (grep tainted($_),
+		 do { local ($@, $SIG{__DIE__}); eval { values %$_ }}) {
+		my (undef, undef, undef, $subn) = caller 1;
+		my $msg = ( $subn =~ /::([a-z]\w*)$/
+			    ? "Insecure argument on '$1' method call"
+			    : "Insecure argument on method call" );
+		_tcroak($msg);
             }
         }
     }
 }
 
 sub _gen_dos2unix {
+    my $unix2dos = shift;
+    my $name = ($unix2dos ? 'unix2dos' : 'dos2unix');
     my $previous;
     my $done;
     sub {
-        $done and die "Internal error: bad calling sequence for unix2dos transformation";
+        $done and die "Internal error: bad calling sequence for $name transformation";
         my $adjustment = 0;
         for (@_) {
             if ($debug and $debug & 128) {
-                _debug ("before dos2unixunix2dos: previous: $previous, data follows...");
+                _debug ("before $name: previous: $previous, data follows...");
                 _hexdump($_);
             }
             if (length) {
@@ -248,7 +245,12 @@ sub _gen_dos2unix {
                     $_ = "\x0d$_";
                 }
                 $adjustment -= $previous = s/\x0d\z//s;
-                $adjustment -= s/\x0d\x0a/\x0a/gs;
+                if ($unix2dos) {
+                    $adjustment += s/(?<!\x0d)\x0a/\x0d\x0a/gs;
+                }
+                else {
+                    $adjustment -= s/\x0d\x0a/\x0a/gs;
+                }
             }
             elsif ($previous) {
                 $previous = 0;
@@ -257,7 +259,7 @@ sub _gen_dos2unix {
                 $_ = "\x0d";
             }
             if ($debug and $debug & 128) {
-                _debug ("after dos2unix: previous: $previous, adjustment: $adjustment, data follows...");
+                _debug ("after $name: previous: $previous, adjustment: $adjustment, data follows...");
                 _hexdump($_);
             }
             return $adjustment;
@@ -265,21 +267,6 @@ sub _gen_dos2unix {
     }
 }
 
-sub _unix2dos {
-    if ($debug and $debug & 128) {
-        _debug ("before unix2dos: data follows...");
-        _hexdump($_[0]);
-    }
-    my $adjustment = $_[0] =~ s/\x0a/\x0d\x0a/gs;
-    if ($debug and $debug & 128) {
-        _debug ("before unix2dos: adjustment: $adjustment, data follows...");
-        _hexdump($_[0]);
-    }
-    $adjustment;
-}
-
-sub _gen_unix2dos { \&_unix2dos }
-
 sub _gen_converter {
     my $conversion = shift;
 
@@ -298,15 +285,19 @@ sub _gen_converter {
         }
     }
     elsif ($conversion eq 'dos2unix') {
-        return _gen_dos2unix;
+        return _gen_dos2unix(0);
     }
     elsif ($conversion eq 'unix2dos') {
-        return _gen_unix2dos
+        return _gen_dos2unix(1);
     }
     else {
         croak "unknown conversion '$conversion'";
     }
 }
 
+sub _is_lnk { (0120000 & shift) == 0120000 }
+sub _is_dir { (0040000 & shift) == 0040000 }
+sub _is_reg { (0100000 & shift) == 0100000 }
+
 1;
 
